/*
 *   timer.c
 *
 *   This file is part of Emu48
 *
 *   Copyright (C) 1995 Sebastien Carlier
 *
 */
#include "pch.h"
#include "Emu48.h"
#define UNIX_0_TIME	0x0001cf2e8f800000

// Ticks for 'auto off'
#define OFF_TIME	((LONGLONG) (AUTO_OFF * 60) << 13)

// memory address for clock and auto off
// S(X) = 0x70052-0x70070, G(X) = 0x80058-0x80076, 49G = 0x80058-0x80076
#define RPLTIME		((cCurrentRomType=='S')?0x52:0x58)


#define AUTO_OFF    10						// Time in minutes for 'auto off'

BOOL bAccurateTimer = TRUE;
UINT uT1Period  = 62;

static BOOL  bStarted   = FALSE;
static UINT  uT1TimerId = 0;
static UINT  uT2TimerId = 0;
static DWORD dwT1Ticks = 0;
static DWORD dwT2Init  = 0;
static DWORD dwT2Step  = 0;
static DWORD dwT2Ticks = 0;

static __inline MAX(int a, int b) {return (a>b)?a:b;} 

#ifdef UNDER_CE
static void CALLBACK TimeProc(HWND hwnd, UINT uMsg, UINT uEventID, DWORD dwTime);
#else
static void CALLBACK TimeProc(UINT uEventId, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2);
#endif // UNDER_CE

static VOID CheckT1()
{
#ifdef UNDER_CE
	if (!bStarted)
		return;
#else // UNDER_CE
	_ASSERT(bStarted);
#endif // UNDER_CE
	if ((Chipset.t1&8) == 0)
	{
		Chipset.IORam[0x2E] &= 0x7; // clear bit 3
		return;
	}
	if (Chipset.IORam[0x2E]&4)
	{ // T1 -> Wake Up
		if (Chipset.Shutdn)
		{
			ResumeThread(hThread);
			Chipset.IORam[0x2E] &= 0xB; // clear bit 2
			Chipset.IORam[0x2E] |= 0x8; // set bit 3
		}
	}
	if (Chipset.IORam[0x2E]&2)
	{ // T1 -> Interrupt
		Chipset.IORam[0x2E] |= 0x8; // set bit 3
		Chipset.SoftInt = TRUE;
		bInterrupt=TRUE;
	}
	return;
}

static VOID CheckT2()
{
#ifdef UNDER_CE
	if (!bStarted)
		return;
#else // UNDER_CE
	_ASSERT(bStarted);
#endif // UNDER_CE
	if ((Chipset.t2&0x80000000) == 0)
	{
		Chipset.IORam[0x2F] &= 7; // clear bit 3
		return;
	}
	if (Chipset.IORam[0x2F]&4)
	{ // T2 -> Wake Up
		if (Chipset.Shutdn)
		{
			ResumeThread(hThread);
			Chipset.IORam[0x2F] &= 0xB; // clear bit 2
			Chipset.IORam[0x2F] |= 0x8; // set bit 3
		}
	}
	if (Chipset.IORam[0x2F]&2)
	{ // T2 -> Interrupt
		Chipset.IORam[0x2F] |= 0x8; // set bit 3
		Chipset.SoftInt = TRUE;
		bInterrupt=TRUE;
	}
	return;
}

static VOID RescheduleT2()
{
#ifdef UNDER_CE
	if (uT2TimerId)
		KillTimer(hwndOwner, uT2TimerId);
#else // UNDER_CE
	_ASSERT(uT2TimerId == 0);
#endif // UNDER_CE
	dwT2Init   = GetTickCount();
	dwT2Ticks  = dwT2Init;
#ifdef UNDER_CE
	uT2TimerId = SetTimer(hwndOwner, 2, MAX(16,Chipset.t2>>3), (TIMERPROC)&TimeProc);
#else // UNDER_CE
	uT2TimerId = timeSetEvent(MAX(16,Chipset.t2>>3),0,(LPTIMECALLBACK)&TimeProc,2,TIME_ONESHOT);
#endif // UNDER_CE
	return;
}

static VOID AbortT2()
{
#ifdef UNDER_CE
	if (!uT2TimerId)
		return;
#else // UNDER_CE
	_ASSERT(uT2TimerId != 0);
#endif // UNDER_CE
	uT2TimerId = 0;
#ifdef UNDER_CE
	KillTimer(hwndOwner, uT2TimerId);
#else // UNDER_CE
	timeKillEvent(uT2TimerId);
#endif // UNDER_CE
	return;
}

#ifdef UNDER_CE
static void CALLBACK TimeProc(HWND hwnd, UINT uMsg, UINT uEventId, DWORD dwTime)
{
	if (!bStarted)
		return;
	if ((uT1TimerId!=0) && (uEventId == uT1TimerId))
	{
		Chipset.t1 = (Chipset.t1-1)&0xF;
		CheckT1();
		return;
	}
	if ((uT2TimerId!=0) && (uEventId == uT2TimerId))
	{
		uT2TimerId = 0;
		Chipset.t2 = 0xFFFFFFFF;
		CheckT2();
		RescheduleT2();
		return;
	}
	return;
	UNREFERENCED_PARAMETER(uMsg);
	UNREFERENCED_PARAMETER(dwTime);
}
#else // UNDER_CE
static void CALLBACK TimeProc(UINT uEventId, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	if (!bStarted)
		return;
	if ((uT1TimerId!=0) && (uEventId == uT1TimerId))
	{
		Chipset.t1 = (Chipset.t1-1)&0xF;
		CheckT1();
		return;
	}
	if ((uT2TimerId!=0) && (uEventId == uT2TimerId))
	{
		uT2TimerId = 0;
		Chipset.t2 = 0xFFFFFFFF;
		CheckT2();
		RescheduleT2();
		return;
	}
	return;
	UNREFERENCED_PARAMETER(uMsg);
}
#endif // UNDER_CE


VOID StartTimers()
{
	if (bStarted)
		return;
	if (Chipset.IORam[0x2F]&1)
	{
		bStarted = TRUE;
#ifndef UNDER_CE
		if (bAccurateTimer)
		{
			MMRESULT uResult;
			uResult = timeBeginPeriod(1);
			if (uResult!=TIMERR_NOERROR)
				bAccurateTimer=FALSE;
		}
		uT1TimerId = timeSetEvent(uT1Period,0,(LPTIMECALLBACK)&TimeProc,1,TIME_PERIODIC);
#else // UNDER_CE
		uT1TimerId = SetTimer(hwndOwner, 1, uT1Period, (TIMERPROC)&TimeProc);
#endif // UNDER_CE
		RescheduleT2();
	}
	return;
}

VOID StopTimers()
{
	if (!bStarted)
		return;
	Chipset.t2 = ReadT2();
	if (uT1TimerId != 0)
	{
#ifdef UNDER_CE
		KillTimer(hwndOwner, uT1TimerId);
#else // UNDER_CE
		timeKillEvent(uT1TimerId);
#endif // UNDER_CE
		uT1TimerId = 0;
	}
	if (uT2TimerId != 0)
		AbortT2();
	bStarted = FALSE;
#ifndef UNDER_CE
	if (bAccurateTimer)
	{
		timeEndPeriod(1);
	}
#endif // UNDER_CE
	return;
}

DWORD ReadT2()
{
	DWORD dwT2;
	if (!bStarted)
	{
		dwT2 = Chipset.t2;
	}
	else
	{
		DWORD dwNow = GetTickCount();
		if (dwNow == dwT2Ticks)
		{
			if (dwT2Step < 7)
				dwT2Step++;
			return Chipset.t2-(8*(dwNow-dwT2Init)+dwT2Step);
		}
		dwT2Step = 0;
		dwT2Ticks = dwNow;
		dwT2 = Chipset.t2-8*(dwNow-dwT2Init);
	}
	return dwT2;
}

VOID SetT2(DWORD dwValue)
{
	if (!bStarted)
	{
		Chipset.t2 = dwValue;
		return;
	}
	if (uT2TimerId != 0)
		AbortT2();
	Chipset.t2 = dwValue;
	CheckT2();
	RescheduleT2();
	return;
}

BYTE ReadT1()
{
	return Chipset.t1;
}

VOID SetT1(BYTE byValue)
{
	Chipset.t1 = byValue&0xF;
	CheckT1();
	return;
}

VOID SetHP48Time(VOID)						// set date and time
{
	SYSTEMTIME ts;
	LONGLONG   ticks, time;
	DWORD      dw;
	WORD       crc, i;
	BYTE       p[4];

	_ASSERT(sizeof(LONGLONG) == 8);			// check size of datatype

	if (!Chipset.Port0)
		return;

	GetLocalTime(&ts);						// local time, _ftime() cause memory/resource leaks

	// calculate days until 01.01.1970
	dw = (DWORD) ts.wMonth;
	if (dw > 2)
		dw -= 3L;
	else
	{
		dw += 9L;
		--ts.wYear;
	}
	dw = (DWORD) ts.wDay + (153L * dw + 2L) / 5L;
	dw += (146097L * (((DWORD) ts.wYear) / 100L)) / 4L;
	dw +=   (1461L * (((DWORD) ts.wYear) % 100L)) / 4L;
	dw -= 719469L;

	// convert into seconds and add time
	dw = dw * 24L + (DWORD) ts.wHour;
	dw = dw * 60L + (DWORD) ts.wMinute;
	dw = dw * 60L + (DWORD) ts.wSecond;

	// create timerticks = (s + ms) * 8192
	ticks = ((LONGLONG) dw << 13) | (((LONGLONG) ts.wMilliseconds << 10) / 125);

	ticks += UNIX_0_TIME;					// add offset ticks from year 0
	ticks += (LONG) Chipset.t2;				// add actual timer2 value

	time = ticks;							// save for calc. timeout
	time += OFF_TIME;						// add 10 min for auto off

	dw = RPLTIME;							// HP addresses for clock in port0

	crc = 0x0;								// reset crc value
	for (i = 0; i < 13; ++i, ++dw)			// write date and time
	{
		*p = (BYTE) ticks & 0xf;
		crc = (crc >> 4) ^ (((crc ^ ((WORD) *p)) & 0xf) * 0x1081);
		Chipset.Port0[dw] = *p;				// always store in port0
		ticks >>= 4;
	}

	Nunpack(p,crc,4);						// write crc
	memcpy(Chipset.Port0+dw,p,4);			// always store in port0

	dw += 4;								// HP addresses for timeout

	for (i = 0; i < 13; ++i, ++dw)			// write time for auto off
	{
		// always store in port0
		Chipset.Port0[dw] = (BYTE) time & 0xf;
		time >>= 4;
	}

	Chipset.Port0[dw] = 0xf;				// always store in port0
	return;
}
